@@ -4,7 +4,7 @@
  * Part of icu_ext: a PostgreSQL extension to expose functionality from ICU
  * (see http://icu-project.org)
  *
- * By Daniel Vérité, 2018-2019. See LICENSE.md
+ * By Daniel Vérité, 2018-2020. See LICENSE.md
  */
 
 #include "postgres.h"
@@ -12,6 +12,7 @@
 #include "catalog/pg_collation.h"
 #include "fmgr.h"
 #include "funcapi.h"
+#include "lib/stringinfo.h"
 #include "miscadmin.h"
 #include "mb/pg_wchar.h"
 #include "utils/builtins.h"
@@ -22,6 +23,7 @@
 #include "unicode/ucol.h"
 #include "unicode/uloc.h"
 #include "unicode/umachine.h"
+#include "unicode/uscript.h"
 #include "unicode/ustring.h"
 #include "unicode/utext.h"
 #include "unicode/uvernum.h"
@@ -109,7 +111,7 @@ icu_collation_attributes(PG_FUNCTION_ARGS)
 
 	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
 		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),	
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 			 errmsg("set-valued function called in context that cannot accept a set")));
 
 	/* Switch into long-lived context to construct returned data structures */
@@ -139,7 +141,8 @@ icu_collation_attributes(PG_FUNCTION_ARGS)
 	/* name (not a real attribute, added for convenience) */
 	if (include_defaults)
 	{
-		UChar dname_local[100];
+		/* Use a large initial buffer to avoid bug ICU-21157 */
+		UChar dname_local[500];
 		UChar *dname = dname_local;
 		char *buf;
 		int32_t ulen;
@@ -147,7 +150,7 @@ icu_collation_attributes(PG_FUNCTION_ARGS)
 		ulen = uloc_getDisplayName(locale,
 								   NULL,
 								   dname,
-								   sizeof(dname_local)-1,
+								   sizeof(dname_local)/sizeof(UChar),
 								   &status);
 		if (status == U_BUFFER_OVERFLOW_ERROR)
 		{
@@ -282,6 +285,109 @@ icu_collation_attributes(PG_FUNCTION_ARGS)
 		tuplestore_putvalues(tupstore, tupdesc, values, nulls);
 	}
 
+	/* Max variable (key:kv) */
+	{
+		UColReorderCode reorder_code = ucol_getMaxVariable(collator);
+		const char *code_name = NULL;
+		switch(reorder_code)
+		{
+			case UCOL_REORDER_CODE_SPACE:
+				code_name = "space";
+				break;
+			case UCOL_REORDER_CODE_PUNCTUATION:
+				code_name = "punct";
+				break;
+			case UCOL_REORDER_CODE_SYMBOL:
+				code_name = "symbol";
+				break;
+			case UCOL_REORDER_CODE_CURRENCY:
+				code_name = "currency";
+				break;
+			case UCOL_REORDER_CODE_DIGIT:
+				code_name = "digit";
+				break;
+			default:
+				break;
+		}
+		/* "punct" is the default. Omit it unless include_defaults is set */
+		if (code_name != NULL && (include_defaults ||
+					  reorder_code != UCOL_REORDER_CODE_PUNCTUATION))
+		{
+			values[0] = CStringGetTextDatum("kv");
+			values[1] = CStringGetTextDatum(code_name);
+			tuplestore_putvalues(tupstore, tupdesc, values, nulls);
+		}
+	}
+
+	/* Reorder codes (key:kr) */
+	{
+		UErrorCode status = U_ZERO_ERROR;
+		StringInfoData aggr_values;  /* 4-letter codes separated by hyphens */
+		int32_t *reorder_codes = NULL;
+		int32_t nb_reorderings = ucol_getReorderCodes(collator,
+												NULL,
+												0,
+												&status);
+		if (status != U_BUFFER_OVERFLOW_ERROR && U_FAILURE(status))
+			elog(ERROR, "uloc_getReorderCodes failed: %s", u_errorName(status));
+
+		initStringInfo(&aggr_values);
+
+		if (nb_reorderings > 0)
+		{
+			reorder_codes = palloc(nb_reorderings*sizeof(int32_t));
+			status = U_ZERO_ERROR;
+			nb_reorderings = ucol_getReorderCodes(collator,
+												  reorder_codes,
+												  nb_reorderings,
+												  &status);
+			if (status != U_BUFFER_OVERFLOW_ERROR && U_FAILURE(status))
+				elog(ERROR, "uloc_getReorderCodes failed: %s", u_errorName(status));
+		}
+
+		for (uint32_t idx=0; idx < nb_reorderings; idx++)
+		{
+			const char *value = NULL;
+			if (reorder_codes[idx] >= UCOL_REORDER_CODE_FIRST)
+			{
+				switch(reorder_codes[idx])
+				{
+					case UCOL_REORDER_CODE_SPACE:
+						value = "space";
+						break;
+					case UCOL_REORDER_CODE_PUNCTUATION:
+						value = "punct";
+						break;
+					case UCOL_REORDER_CODE_SYMBOL:
+						value = "symbol";
+						break;
+					case UCOL_REORDER_CODE_CURRENCY:
+						value = "currency";
+						break;
+					case UCOL_REORDER_CODE_DIGIT:
+						value = "digit";
+						break;
+				}
+			}
+			else
+			{
+				value  = uscript_getShortName((UScriptCode)reorder_codes[idx]);
+			}
+			if (value != NULL)
+			{
+				if (idx >= 1)
+					appendStringInfoChar(&aggr_values, '-');
+				appendStringInfoString(&aggr_values, value);
+			}
+		}
+		if (aggr_values.len > 0)
+		{
+			values[0] = CStringGetTextDatum("kr");
+			values[1] = CStringGetTextDatum(aggr_values.data);
+			tuplestore_putvalues(tupstore, tupdesc, values, nulls);
+		}
+	}
+
 	/* version (not a real attribute, added for convenience) */
 	if (include_defaults)
 	{
@@ -335,7 +441,7 @@ icu_locales_list(PG_FUNCTION_ARGS)
 
 	if (rsinfo == NULL || !IsA(rsinfo, ReturnSetInfo))
 		ereport(ERROR,
-				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),	
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 			 errmsg("set-valued function called in context that cannot accept a set")));
 
 	/* Switch into long-lived context to construct returned data structures */
@@ -431,7 +537,7 @@ icu_locales_list(PG_FUNCTION_ARGS)
 			default:
 				layout = ""; break;
 			}
-			
+
 			col_num = add_string(layout, col_num, values, nulls);
 		}
 
@@ -566,7 +672,7 @@ Datum
 icu_compare_coll(PG_FUNCTION_ARGS)
 {
 	text *txt1 = PG_GETARG_TEXT_PP(0);
-	text *txt2 = PG_GETARG_TEXT_PP(1);	
+	text *txt2 = PG_GETARG_TEXT_PP(1);
 	const char *collname = text_to_cstring(PG_GETARG_TEXT_P(2));
 	UCollator	*collator = NULL;
 	UErrorCode	status = U_ZERO_ERROR;
@@ -593,7 +699,7 @@ Datum
 icu_compare(PG_FUNCTION_ARGS)
 {
 	text *txt1 = PG_GETARG_TEXT_PP(0);
-	text *txt2 = PG_GETARG_TEXT_PP(1);	
+	text *txt2 = PG_GETARG_TEXT_PP(1);
 	UCollator *collator = ucollator_from_coll_id(PG_GET_COLLATION());
 	UCollationResult result;
 
@@ -688,7 +794,7 @@ icu_sort_key_coll(PG_FUNCTION_ARGS)
 
 	ulen = icu_to_uchar(&ustring, VARDATA_ANY(txt), VARSIZE_ANY_EXHDR(txt));
 
- 	collator = ucol_open(locname, &status);
+	collator = ucol_open(locname, &status);
 	if (!collator)
 		elog(ERROR, "failed to open collation");
 
